==> fipepy-1.0.4.tar.gz: fipepy-1.0.4/fipe/feature/variables/feature.py <==
from collections.abc import Mapping

import gurobipy as gp
import pandas as pd

from ...mip import MIP, BaseVar
from ...typing import Categories, FeatureType, MNumber, Number, SNumber
from .binary import BinaryVar
from .categorical import CategoricalVar
from .continuous import ContinuousVar

FeatureVar = BinaryVar | ContinuousVar | CategoricalVar
Args = MNumber | Categories


class FeatureVars(BaseVar[SNumber], dict[str, FeatureVar]):
    CLASSES: Mapping[str, type[FeatureVar]] = {
        FeatureType.BIN.value: BinaryVar,
        FeatureType.CAT.value: CategoricalVar,
        FeatureType.CON.value: ContinuousVar,
    }

    def __init__(self, name: str = "") -> None:
        BaseVar.__init__(self, name=name)
        dict.__init__(self)

    def build(self, mip: MIP) -> None:
        for var in self.values():
            var.build(mip)

    def add_var(
        self,
        feature: str,
        vtype: str,
        *,
        levels: MNumber | None = None,
        categories: Categories | None = None,
    ) -> None:
        self[feature] = self._add_var(
            vtype=vtype,
            name=feature,
            levels=levels,
            categories=categories,
        )

    def _apply(self, prop_name: str) -> SNumber:
        series = pd.Series()
        for name, var in self.items():
            value = getattr(var, prop_name)
            if isinstance(value, pd.Series):
                series = pd.concat([series, value])
            else:
                series[name] = value
        return series

    @staticmethod
    def _build_args(**kwargs: Args) -> dict[str, Args]:
        return {
            key: value for key, value in kwargs.items() if value is not None
        }

    @staticmethod
    def _add_var(
        vtype: str,
        *,
        name: str = "",
        levels: MNumber | None = None,
        categories: Categories | None = None,
    ) -> FeatureVar:
        cls = FeatureVars.CLASSES.get(vtype)
        if cls is None:
            msg = f"Invalid variable type: {vtype}"
            raise ValueError(msg)
        args = FeatureVars._build_args(
            levels=levels,
            categories=categories,
        )
        if vtype == FeatureType.CON.value:
            args["floor"] = True

        return cls(name=name, **args)

    @staticmethod
    def fetch(
        feature_var: FeatureVar,
        *,
        level: Number | None = None,
        category: str | None = None,
    ) -> gp.Var | gp.MVar:
        if isinstance(feature_var, BinaryVar):
            return FeatureVars.fetch_bin(feature_var=feature_var)
        if isinstance(feature_var, ContinuousVar):
            return FeatureVars.fetch_con(
                feature_var=feature_var,
                level=level,
            )
        if isinstance(feature_var, CategoricalVar):
            return FeatureVars.fetch_cat(
                feature_var=feature_var,
                category=category,
            )
        msg = f"Unknown feature var type: {type(feature_var)}"
        raise ValueError(msg)

    @staticmethod
    def fetch_bin(feature_var: BinaryVar) -> gp.Var:
        return feature_var.var

    @staticmethod
    def fetch_con(
        feature_var: ContinuousVar,
        level: Number | None,
    ) -> gp.MVar:
        if level is None:
            msg = "Level must be provided for continuous variable."
            raise ValueError(msg)
        j = int(feature_var.levels.searchsorted(level))
        return feature_var[j]

    @staticmethod
    def fetch_cat(
        feature_var: CategoricalVar,
        category: str | None,
    ) -> gp.Var:
        if category is None:
            msg = "Category must be provided for categorical variable."
            raise ValueError(msg)
        return feature_var[category]

==> fipepy-1.0.5.tar.gz: fipepy-1.0.5/fipe/feature/variables/feature.py <==
from collections.abc import Mapping
from typing import override

import gurobipy as gp
import pandas as pd

from ...mip import MIP, BaseVar
from ...typing import Categories, FeatureType, MNumber, Number, SNumber
from .binary import BinaryVar
from .categorical import CategoricalVar
from .continuous import ContinuousVar

FeatureVar = BinaryVar | ContinuousVar | CategoricalVar
Args = MNumber | Categories


class FeatureVars(BaseVar[SNumber], dict[str, FeatureVar]):
    CLASSES: Mapping[str, type[FeatureVar]] = {
        FeatureType.BIN.value: BinaryVar,
        FeatureType.CAT.value: CategoricalVar,
        FeatureType.CON.value: ContinuousVar,
    }

    def __init__(self, name: str = "") -> None:
        BaseVar.__init__(self, name=name)
        dict.__init__(self)

    @override
    def build(self, mip: MIP) -> None:
        for var in self.values():
            var.build(mip)

    def add_var(
        self,
        feature: str,
        vtype: str,
        *,
        levels: MNumber | None = None,
        categories: Categories | None = None,
    ) -> None:
        self[feature] = self._add_var(
            vtype=vtype,
            name=feature,
            levels=levels,
            categories=categories,
        )

    @override
    def _apply(self, prop_name: str) -> SNumber:
        series = pd.Series()
        for name, var in self.items():
            value = getattr(var, prop_name)
            if isinstance(value, pd.Series):
                series = pd.concat([series, value])
            else:
                series[name] = value
        return series

    @staticmethod
    def _build_args(**kwargs: Args) -> dict[str, Args]:
        return {
            key: value for key, value in kwargs.items() if value is not None
        }

    @staticmethod
    def _add_var(
        vtype: str,
        *,
        name: str = "",
        levels: MNumber | None = None,
        categories: Categories | None = None,
    ) -> FeatureVar:
        match vtype:
            case FeatureType.BIN.value:
                return BinaryVar(name=name)
            case FeatureType.CAT.value:
                if categories is None:
                    msg = (
                        "Categories must be provided for categorical variable."
                    )
                    raise ValueError(msg)
                return CategoricalVar(name=name, categories=categories)
            case FeatureType.CON.value:
                if levels is None:
                    msg = "Levels must be provided for continuous variable."
                    raise ValueError(msg)
                return ContinuousVar(name=name, levels=levels, floor=True)
            case _:
                msg = f"Invalid variable type: {vtype}"
                raise ValueError(msg)

    @staticmethod
    def fetch(
        var: FeatureVar,
        *,
        level: Number | None = None,
        category: str | None = None,
    ) -> gp.Var | gp.MVar:
        if isinstance(var, BinaryVar):
            return FeatureVars._fetch_bin(var=var)
        if isinstance(var, ContinuousVar):
            return FeatureVars._fetch_con(var=var, level=level)
        if isinstance(var, CategoricalVar):
            return FeatureVars._fetch_cat(var=var, category=category)
        msg = f"Unknown feature var type: {type(var)}"
        raise ValueError(msg)

    @staticmethod
    def _fetch_bin(var: BinaryVar) -> gp.Var:
        return var.var

    @staticmethod
    def _fetch_con(var: ContinuousVar, level: Number | None) -> gp.MVar:
        if level is None:
            msg = "Level must be provided for continuous variable."
            raise ValueError(msg)
        j = int(var.levels.searchsorted(level))
        return var[j]

    @staticmethod
    def _fetch_cat(var: CategoricalVar, category: str | None) -> gp.Var:
        if category is None:
            msg = "Category must be provided for categorical variable."
            raise ValueError(msg)
        return var[category]
